import transformers


class ModelCardConfig():
    """ This is the configuration class to store the configuration 
    of a `Model Card`. It is used to instantiate model card according
    to the specified arguments, describing the model's development
    and performances. """
    
